/*
 * The Teleworld Internal Kernel Debugger
 * Copyright (c) 1998 Teleworld, Inc. All Rights Reserved
 * Original modifications by Jim Barton, jmb@teleworld.com
 *
 * Portions derived from:
 *
 * >> Routines providing a simple monitor for use on the PowerMac.
 * >>
 * >> Copyright (C) 1996 Paul Mackerras.
 */
#include <linux/errno.h>
#include <linux/sched.h>
#include <asm/ptrace.h>
#include <asm/ppcops.h>
#include <linux/string.h>
#include <asm/mmu.h>
#include <asm/delay.h>
#undef NULL
#include "nonstdio.h"
#include "privinst.h"

#undef XMON_DEBUG

/* Warning: the sense of this define is backwards. For most cases,
 * it should be defined! */
#define NO_RISCWATCH 

#ifdef XMON_DEBUG
static int	ofInterest = 0;
#endif

u_int bus_error_jmp[100];
struct pt_regs _xdebug_faultregs;
#define setjmp xmon_setjmp
#define longjmp xmon_longjmp

/* Breakpoint stuff */
struct bpt {
	unsigned address;
	unsigned instr;
	unsigned count;
	unsigned char enabled;
	int (*condition)(struct pt_regs *regs, struct bpt *bp);
	unsigned cond_data;
	char autodel;
	char sstep;
	int rw;
	int size;
};

#define NBPTS	16
#define NHW	2
static struct bpt bpts[NBPTS];
static struct bpt dabr[NHW];
static struct bpt iabr[NHW];
static unsigned bpinstr = 0x7fe00008;	/* trap */
static char singlestep = 0;

#define IFORMMASK	((0xfc << 26) | (0xcff << 1))
#define BC		(16 << 26)
#define BCCTR		((19 << 26) | (528 << 1))
#define BCLR		((19 << 26) | (16 << 1))
#define ISSUBR(a)	(((a)&0x1) && \
			 ((IFORMMASK & (a)) == BCCTR || \
			  (IFORMMASK & (a)) == BCLR || \
			  ((a) & (0xfc << 26)) == BC))

/* Prototypes */
void (*xmon_fault_handler)(struct pt_regs *) = 0;
void xmon(struct pt_regs *excp);
unsigned long strtoul(char *, char **, int);
void print_address(unsigned);
void print_instr(unsigned int adr, unsigned int instr);
int xos(int ac, char **av, struct pt_regs *regs);
int s403_poll(void);
static int mread(unsigned, void *, int);
static int mwrite(unsigned, void *, int);
void handle_fault(struct pt_regs *);
static void prdump(unsigned, int);
static int ppc_inst_dump(unsigned, int);
void excprint(struct pt_regs *);
void prregs(struct pt_regs *);
extern void checkpte(unsigned long, struct pt_regs *);
static unsigned read_spr(int);
static void write_spr(int, unsigned);
static unsigned read_dcr(int);
static void write_dcr(int, unsigned);
static void remove_bpts(struct pt_regs *);
static void insert_bpts(struct pt_regs *);
int getsp(void);
void backtrace(unsigned int);

extern int print_insn_big_powerpc(FILE *, unsigned long, unsigned);
extern void printf(const char *fmt, ...);
extern int putchar(int ch);
extern int setjmp(u_int *);
extern void longjmp(u_int *, int);

extern void TrapException(struct pt_regs *regs);


/*
 * New hooks for integrating this into our 403 Linux port
 */
#define XDEBUG_ONENTRY		0x1
#define XDEBUG_ONEXC		0x2
#define XDEBUG_BOOTSTOP		0x8

int _xdebug = XDEBUG_ONEXC|XDEBUG_ONENTRY;

static int _xdebug_inxmon = 0;
static int _xdebug_trace = 0;
static int _xdebug_stopafter = 0;
static int _xdebug_kdebug = 0;		/* debug event goes to kernel */
static volatile int _xdebug_debugtrap = 0;       /* trap instruction was deliberate */
struct mm_struct *_xdebug_restore_mm = 0;
void _xdebug_entry(void);

void _exc_debug(struct pt_regs *);
struct {
	short	trap;
        char    report;
	char	stop;
	char	*name;
	char	*description;
} _exc_table[] = {
    { 0x100,  0, 0, "ci",         "critical interrupt pin" },
    { 0x200,  0, 0, "mcheck",     "machine check" },
    { 0x300,  0, 0, "dstore",     "data storage exception" },
    { 0x400,  0, 0, "istore",     "instruction storage exception" },
    { 0x500,  0, 0, "exti",	  "external interrupt" },
    { 0x600,  0, 0, "align",      "alignment exception" },
    { 0x700,  0, 0, "pcheck",     "program check" },
    { 0xc00,  0, 0, "syscall",    "system call" },
    { 0x1000, 0, 0, "pit",        "programmable interval timer" },
    { 0x1010, 0, 0, "fit",        "fixed interval timer" },
    { 0x1020, 0, 0, "wdog",	  "watchdog timer" },
    { 0x1600, 0, 0, "itlb",	  "instruction TLB can't refill" },
    { 0x1700, 0, 0, "dtlb",	  "data TLB can't refill" },
    { 0x1800, 0, 0, "xmon",	  "tcdmon data fault" },
    { 0x2000, 0, 0, "debug",	  "debug exception" },
    { 0 }
};

static void
_exc_mcheck_print(struct thread_struct *tss) {
	printf("Machine Check @0x%x: ", tss->bear);
	if (tss->besr & 0x80000000)
		printf("DATA ");
	else
		printf("IFETCH ");
	if (tss->besr & 0x40000000)
		printf("DMA ");
	if (tss->besr & 0x20000000)
		printf("READ ");
	else
		printf("WRITE ");
	switch (tss->besr & 0x1c000000) {
		case 0: printf("PROTECTION "); break;
		case 1: printf("PARITY "); break;
		case 2: printf("NOTCONFIGURED "); break;
		case 4: printf("BUS  "); break;
		case 6: printf("BUSTIMEOUT  "); break;
	}
	printf("\n");
}

void
xreport(int idx, struct pt_regs *regs) {
	if (!_exc_table[idx].report)
		return;
	if (regs->trap == 0xc00 &&
	    _exc_table[idx].report != -1 &&
	    _exc_table[idx].report != current->pid)
		return;

	printf("%s: pc %p task %p mmpid %d ", _exc_table[idx].name, regs->nip,
		current, mfspr(SPRN_PID));
	switch (regs->trap) {
	case 0x200:
		_exc_mcheck_print((struct thread_struct *) mfspr(SPRN_SPRG3));
		break;
	case 0x600:
		printf("dear %p ", mfspr(SPRN_DEAR));
		break;
	case 0x700:
		printf("esr 0x%x bear %p desr 0x%x ",
			mfspr(SPRN_ESR), mfspr(SPRN_BEAR), mfspr(SPRN_BESR));
		if (mfspr(SPRN_ESR) & 0x08000000)
			printf("<fp> ");
		break;
	case 0x400:
		printf("esr 0x%x ", mfspr(SPRN_ESR));
		break;
	case 0x300:
		printf("esr 0x%x dear %p ", mfspr(SPRN_ESR), mfspr(SPRN_DEAR));
		break;
	case 0x1800:
	case 0x1700:
		printf("esr 0x%x dear %p *pte %p", mfspr(SPRN_ESR),
			mfspr(SPRN_DEAR), *((pte_t *)current->tss.lastpte));
		break;
	case 0x1600:
		printf("*pte %p", *((pte_t *)current->tss.lastpte));
		break;
	case 0x500:
		printf("exier 0x%x exisr 0x%x ", mfdcr(SPRN_EXIER), mfdcr(SPRN_EXISR));
		break;
	case 0xc00: {
		if (regs->gpr[0] == 0x7777)
			printf("srt");
		else if (regs->gpr[0] >= 189)
			printf("?%3d", regs->gpr[0]);
		else
			printf("%d", regs->gpr[0]);
		printf(": 0x%08lx 0x%08lx 0x%08lx ",
			regs->gpr[3], regs->gpr[4], regs->gpr[5]);
		break;
	} }
	printf("\n");
}

void
_xdebug_exc(struct pt_regs *regs, unsigned long realhandler, unsigned long retval) {
	int i;
	int c;
	if (regs->trap == 0x100) {
		/* critical interrupt; who knows what state the machine
		   is in. Re-init the serial port, so we think it's sane. */
		void s403_realinit(void);
		void s403_virtualinit(void);
		s403_realinit();
		printf("CRITICAL INTERRUPT! (All bets are off!)\n");
		xmon(regs);
		s403_virtualinit();
/* ??? should go through debug in traps.c. */
		goto goback;
	}

	remove_bpts(regs);
	c = s403_poll();
	if (c != -1) {
		extern int s403_ready;

		if (s403_ready == -2)
			s403_ready = 0;
		printf("Keyboard poll event, entering TCDMON (pre-exception handling)\n");
		xmon(regs);
	}
        if (regs->trap == 0x2000) {
                _exc_debug(regs);
                goto goback;
        }

        if (_xdebug_inxmon) {
		if (!xmon_fault_handler) {
			printf("Debugger fault!\n");
			excprint(regs);
			prregs(regs);
			printf("Have fun debugging the debugger!\n");
			xmon(regs);
		}
		else
                    (*xmon_fault_handler)(regs);
                goto goback;
        }

	for (i = 0; _exc_table[i].trap != 0; i++)
		if (_exc_table[i].trap == regs->trap)
                        break;
        if (_exc_table[i].trap == 0) {
                printf("Unknown exception: 0x%x\n", regs->trap);
                excprint(regs);
                prregs(regs);
                xmon(regs);
                goto goback;
        }

        if (realhandler == 0) {
                printf("Kernel exception 0x%x: %s\n", regs->trap, _exc_table[i].description);
                excprint(regs);
                prregs(regs);
                xmon(regs);
                goto goback;
        }

        xreport(i, regs);
        if (_exc_table[i].stop) {
                switch (regs->trap) {
                case 0xc00:
                    if (_exc_table[i].stop != -1) {
                        if (_exc_table[i].stop == regs->gpr[0]) {
                            printf("Syscall %3d: 0x%08lx 0x%08lx 0x%08lx\n",
                                    regs->gpr[0],
                                    regs->gpr[3], regs->gpr[4], regs->gpr[5]);
                        }
                        else
                            goto goback;
                    }
                }
                printf("Stop requested, task state 0x%x\n",
                        _exc_table[i].name, regs);
                xmon(regs);
        }

    goback:
        if (_xdebug_restore_mm) {
            current->mm = _xdebug_restore_mm;
            mtspr(SPRN_PID, current->mm->context);
	    current->tss.pg_tables = (unsigned long *) current->mm->pgd;
            _xdebug_restore_mm = 0;
        }
	if (realhandler && (regs->trap == _xdebug_stopafter)) {
		extern int xbp(int, char **, struct pt_regs *);
		char addr[16];
		char *xargs[4];
		xargs[0] = "bp";
		xargs[1] = "set";
		xargs[2] = "next";
		xargs[3] = addr;
		sprintf(addr, "0x%lx", retval);
		(void)xbp(4, xargs, regs);
	}
	insert_bpts(regs);
	if (realhandler
	    && !(regs->trap == 0x2000 && !_xdebug_kdebug)
	    ) {
		extern void call_realhandler(struct pt_regs *,	/* state */
					     unsigned long,	/* PC */
					     unsigned long,	/* MSR */
					     unsigned long);	/* LR */
		_xdebug_kdebug = 0;
		call_realhandler(regs, realhandler, MSR_KERNEL, retval);
	}
}

void
debug(char *s, struct pt_regs *regs) {
	unsigned long flags;
	save_flags(flags);
	cli();
        if (regs) {
            printk("Kernel process exception: %s\n", s);
            printk("Exception 0x%lx: NIP: %08lX, MSR: %08lX, FRAME: 0x%p\n",
                   regs->trap, regs->nip, regs->msr, regs);
            printk("ESR 0x%x BESR 0x%x BEAR 0x%x DEAR 0x%x\n",
                    mfspr(SPRN_ESR), mfspr(SPRN_BESR), mfspr(SPRN_BEAR),
                    mfspr(SPRN_DEAR));
            if (current)
                    printk("TASK = %p[%d] '%s'\n", current, current->pid, current->comm);
            xmon(regs);
        }
        else {
            printk("Kernel debug halt requested: %s\n", s);
            _xdebug_debugtrap = 1;
            asm volatile(".long 0x7fe00008");
            _xdebug_debugtrap = 0;
        }
	restore_flags(flags);
}

void
_exc_debug(struct pt_regs *regs) {
	unsigned long dbsr = mfspr(SPRN_DBSR);
	unsigned int check;
	if (user_mode(regs) && 
	    (current && (current->tss.flags & TASK_FLAG_SS))) {
		/* kernel owns this event */
		_xdebug_kdebug = 1;
		return;
	}
	if (dbsr & 0x80000000) {
		/* single step completed */
		singlestep = 0;
		if (_xdebug_trace > 0) {
			_xdebug_trace--;
			singlestep = 1;
		}
	}
	/*
	if (dbsr & 0x80000000) {
		printf("Single step completed.  Didn't belong to kernel, PC 0x%x, dbsr: 0x%x\n",
			regs->nip, dbsr);
	}
	*/
	if (dbsr & 0x04000000) {
		/* IAC1 */
		printf("Instruction compare 1 trap; PC 0x%x\n", regs->nip);
	}
	if (dbsr & 0x02000000) {
		/* IAC2 */
		printf("Instruction compare 2 trap; PC 0x%x\n", regs->nip);
	}
	if (dbsr & 0x01000000) {
		/* DAC1 R */
		printf("Data read compare 1 trap; PC 0x%x\n", regs->nip);
		dabr[0].instr = regs->nip;
	}
	if (dbsr & 0x00800000) {
		/* DAC1 W */
		printf("Data write compare 1 trap; PC 0x%x\n", regs->nip);
		dabr[0].instr = regs->nip;
	}
	if (dbsr & 0x00400000) {
		/* DAC2 R */
		printf("Data read compare 2 trap; PC 0x%x\n", regs->nip);
		dabr[1].instr = regs->nip;
	}
	if (dbsr & 0x00200000) {
		/* DAC2 W */
		printf("Data write compare 2 trap; PC 0x%x\n", regs->nip);
		dabr[1].instr = regs->nip;
	}
	if (dbsr & 0x10000000) {
		/* trap event; might be a breakpoint */
		struct bpt *bp = bpts;
		int i;

		for (i = 0; i < NBPTS; ++i, ++bp) {
			if (!bp->enabled)
				continue;
			if (bp->address == regs->nip) {
				if (bp->condition) {
					if (!bp->condition(regs, bp)) {
					    dbsr &= 0xfff00000;
					    mtspr(SPRN_DBSR, dbsr);
					    return;
					}
				}
				printf("Breakpoint at PC 0x%x", regs->nip);
				if (bp->autodel) {
					printf(", return value 0x%x", regs->gpr[3]);
					bp->enabled = 0;
				}
				break;
			}
		}
		if (i == NBPTS)
		{
			if (user_mode(regs)) {
				/*
				 * It's for the kernel, but we're not on the current
				 * kernel stack. We need to get there ...
				 */
				extern void kdebugtrap(unsigned long,
				                       unsigned long,
				                       int);
				unsigned long ksp = current->tss.ksp -
					((2 * STACK_FRAME_OVERHEAD) + sizeof(struct pt_regs));

				memcpy((void *)ksp+STACK_FRAME_OVERHEAD,
				       regs,
				       sizeof(struct pt_regs));
			        current->tss.last_pc = regs->nip;
			        current->tss.user_stack = regs->gpr[1];
			        current->tss.regs = regs;
				dbsr &= 0xfff00000;
				mtspr(SPRN_DBSR, dbsr);
				kdebugtrap(ksp+STACK_FRAME_OVERHEAD, ksp, MSR_KERNEL);
				/* NOTREACHED (really!) */
			}
                        else if (_xdebug_debugtrap) {
				printf("Kernel requested debug trap, regs at 0x%x",
					regs);

                                /*
                                 * Bump the instruction up one, so we can continue
                                 * from the debug trap.
                                 */
                                regs->nip += 4;
                        }
			else {
				printf("Unexpected kernel breakpoint at PC 0x%x, MSR 0x%x",
					regs->nip, regs->msr);
                        }
		}
		printf("\n");
	}
	dbsr &= 0xfff00000;
	mtspr(SPRN_DBSR, dbsr);
        if (!_xdebug_debugtrap) {
            if (mread(regs->nip, &check, 4) != 4)
                    printf("0x%x, no valid instruction\n", regs->nip);
            else {
                    printf("0x%x ", regs->nip);
                    print_instr(regs->nip, check);
                    printf("\n");
            }
        }
	if (!singlestep)
	    xmon(regs);
}

void
_xdebug_entry() {
	struct pt_regs regs;
#ifdef NO_RISCWATCH
	mtspr(SPRN_DBCR, 0x40000000); /* go into internal debug mode */
#endif
	if (_xdebug & XDEBUG_BOOTSTOP) {
		printf("\rHello, the TCDMON internal debugger has acquired boot control\n");
		asm volatile ("stw	0,0(%0)\n\
			lwz	0,0(1)\n\
			stw	0,4(%0)\n\
			stmw	2,8(%0)" : : "r" (&regs));
		regs.nip = regs.link = ((unsigned long *)regs.gpr[1])[1];
		regs.msr = get_msr();
		regs.ctr = get_ctr();
		regs.xer = get_xer();
		regs.ccr = get_cr();
		regs.trap = 0x2000;
		xmon(&regs);
#ifdef XMON_DEBUG
		printf("calling insert_bpts ...\n");
		ofInterest = 1;
#endif
		insert_bpts(&regs);
#ifdef XMON_DEBUG
		ofInterest = 0;
		printf("setting MSR 0x%x, DBCR 0x%x\n", regs.msr, mfspr(SPRN_DBCR));
#endif
	}
	else
		printf("TCDMON internal debugger activated\n");
	mtmsr(MSR_KERNEL);
}

void
xmon_irq(int irq, void *d, struct pt_regs *regs)
{
	printf("Keyboard interrupt\n");
	xmon(regs);
}

static void
insert_bpts(struct pt_regs *regs)
{
	int i;
	struct bpt *bp;
	unsigned long odbcr;

	bp = bpts;
	for (i = 0; i < NBPTS; ++i, ++bp) {
		if (!bp->enabled)
			continue;
		if (regs->nip == bp->address) {
			bp->sstep = 1;
			continue;
		}
		if (mread(bp->address, &bp->instr, 4) != 4
		    || mwrite(bp->address, &bpinstr, 4) != 4) {
			printf("Couldn't insert breakpoint at %x, disabling\n",
			       bp->address);
			bp->enabled = 0;
		}
#if 0
		if (regs->trap == 0x2000 && bp->enabled) {
			unsigned check;
			mread(bp->address, &check, 4);
			printf("inserted bp @ 0x%x replacing 0x%x (verify 0x%x)\n",
				bp->address, bp->instr, check);
		}
#endif
	}

	odbcr = mfspr(SPRN_DBCR);
	odbcr |= 0x01000000;	/* trap debug event */
	odbcr &= ~0x00f80000;	/* clear event counter */
	odbcr |= 0x00040000;	/* freeze timers */

        if (dabr[0].enabled && (dabr[0].instr != regs->nip)) {
            mtspr(SPRN_DAC1, dabr[0].address);
            switch (dabr[0].rw) {
              case 0: odbcr |= 0x8000; odbcr &= ~0x4000; break;
              case 1: odbcr |= 0x4000; odbcr &= ~0x8000; break;
              case 2: odbcr |= 0xc000; break;
            }
            switch (dabr[0].size) {
              case 0: odbcr &= ~0x3000; break;
              case 1: odbcr &= ~0x3000; odbcr |= 0x1000; break;
              case 2: odbcr &= ~0x3000; odbcr |= 0x2000; break;
              case 3: odbcr |= 0x3000; break;
            }
        }
#ifdef NO_RISCWATCH
	else
		odbcr &= ~0xc000;
#endif
        if (dabr[1].enabled && (dabr[1].instr != regs->nip)) {
            mtspr(SPRN_DAC2, dabr[1].address);
            switch (dabr[1].rw) {
              case 0: odbcr |= 0x800; odbcr &= ~0x400; break;
              case 1: odbcr |= 0x400; odbcr &= ~0x800; break;
              case 2: odbcr |= 0xc00; break;
            }
            switch (dabr[1].size) {
              case 0: odbcr &= ~0x300; break;
              case 1: odbcr &= ~0x300; odbcr |= 0x100; break;
              case 2: odbcr &= ~0x300; odbcr |= 0x200; break;
              case 3: odbcr |= 0x300; break;
            }
        }
#ifdef NO_RISCWATCH
	else
		odbcr &= ~0x300;
#endif
	if (iabr[0].enabled && (iabr[0].address != regs->nip)) {
		odbcr |= 0x20000;
		mtspr(SPRN_IAC1, iabr[0].address);
	}
#ifdef NO_RISCWATCH
	else
		odbcr &= ~0x20000;
#endif
	if (iabr[1].enabled && (iabr[1].address != regs->nip)) {
		odbcr |= 0x10000;
		mtspr(SPRN_IAC2, iabr[1].address);
	}
#ifdef NO_RISCWATCH
	else
		odbcr &= ~0x10000;
#endif
	if (singlestep) {
		odbcr |= 0x08000000;	/* instruction completion event */
		regs->msr &= ~0x00008000;/* don't allow interrupts */
	}
	mtspr(SPRN_DBCR, odbcr);
	regs->msr |= 0x200;		/* debug exception [DE] bit */
}

static void
remove_bpts(struct pt_regs *regs)
{
	int i;
	struct bpt *bp;
	unsigned instr;
	unsigned long odbcr;

	odbcr = mfspr(SPRN_DBCR);
	odbcr &= 0x40000003;
	mtspr(SPRN_DBCR, odbcr);

	bp = bpts;
	for (i = 0; i < NBPTS; ++i, ++bp) {
		if (!bp->enabled)
			continue;
		if (bp->sstep) {
			bp->sstep = 0;
			continue;
		}
		if (mread(bp->address, &instr, 4) != 4)
			printf("couldn't read 0x%x to remove breakpoint\n", bp->address);
		else if (instr != bpinstr) {
			bp->enabled = 0;
			printf("trap at 0x%x has been overwritten\n", bp->address);
		}
		else if (mwrite(bp->address, &bp->instr, 4) != 4)
			printf("overwrite trap at 0x%x failed\n", bp->address);
#if 0
		else if (regs->trap == 0x2000) {
			unsigned check;
			mread(bp->address, &check, 4);
			printf("removed bp @ 0x%x, put back 0x%x (verify 0x%x)\n",
				bp->address, bp->instr, check);
		}
#endif
	}

	regs->msr &= ~0x200;		/* debug exception [DE] bit */
}

int
xregister_immediate(int ac, char ** av, struct pt_regs *regs) {
	if (!strcmp(av[0], "mtmsr")) {
		unsigned long sv;
		if (ac != 2) {
			printf("mtmsr requires the new MSR value\n");
			return 0;
		}
		sv = strtoul(av[1], 0, 0);
		mtmsr(sv);
		return 0;
	}
	if (!strcmp(av[0], "mfmsr")) {
		printf("MSR = 0x%x\n", mfmsr());
		return 0;
	}
	if (!strcmp(av[0], "mtspr")) {
		unsigned long reg;
		unsigned long sv;
		if (ac != 3) {
			printf("mtspr requires the SPR and a value\n");
			return 0;
		}
		reg = strtoul(av[1], 0, 0);
		sv = strtoul(av[2], 0, 0);
		write_spr(reg, sv);
		return 0;
	}
	if (!strcmp(av[0], "mfspr")) {
		unsigned long sv;
		if (ac != 2) {
			printf("mfspr requires the the SPR\n");
			return 0;
		}
		sv = strtoul(av[1], 0, 0);
		printf("SPR 0x%x = 0x%x\n", sv, read_spr(sv));
		return 0;
	}
	if (!strcmp(av[0], "mtdcr")) {
		unsigned long reg;
		unsigned long sv;
		if (ac != 3) {
			printf("mtdcr requires the DCR and a value\n");
			return 0;
		}
		reg = strtoul(av[1], 0, 0);
		sv = strtoul(av[2], 0, 0);
		write_dcr(reg, sv);
		return 0;
	}
	if (!strcmp(av[0], "mfdcr")) {
		unsigned long sv;
		if (ac != 2) {
			printf("mfdcr requires the the DCR\n");
			return 0;
		}
		sv = strtoul(av[1], 0, 0);
		printf("DCR 0x%x = 0x%x\n", sv, read_dcr(sv));
		return 0;
	}
	printf("unknown register command\n");
	return 0;
}

int
xregister_context(int ac, char ** av, struct pt_regs *context) {
	int idx = 1;
	if (ac > 1 && (av[1][0] >= '0' && av[1][0] <= '9')) {
		context = (struct pt_regs *) strtoul(av[1], 0, 0);
		idx++;
	}
	if (ac == idx) {
		excprint(context);
		prregs(context);
		return 0;
	}
	if (!strcmp(av[idx], "mtmsr")) {
		unsigned long sv;
		if (ac < 3) {
			printf("mtmsr requires the new MSR value\n");
			return 0;
		}
		sv = strtoul(av[idx+1], 0, 0);
		context->msr = sv;
		return 0;
	}
	if (!strcmp(av[idx], "mfmsr")) {
		printf("(0x%x) MSR = 0x%x\n", context, context->msr);
		return 0;
	}
	if (!strcmp(av[idx], "mtlr")) {
		unsigned long sv;
		if (ac < 3) {
			printf("mtlr requires the new link value\n");
			return 0;
		}
		sv = strtoul(av[idx+1], 0, 0);
		context->link = sv;
		return 0;
	}
	if (!strcmp(av[idx], "mflr")) {
		printf("(0x%x) LINK = 0x%x\n", context, context->link);
		return 0;
	}
	if (!strcmp(av[idx], "mtctr")) {
		unsigned long sv;
		if (ac < 3) {
			printf("mtctr requires the new counter value\n");
			return 0;
		}
		sv = strtoul(av[idx+1], 0, 0);
		context->ctr = sv;
		return 0;
	}
	if (!strcmp(av[idx], "mfctr")) {
		printf("(0x%x) CTR = 0x%x\n", context, context->ctr);
		return 0;
	}
	if (!strcmp(av[idx], "mtgpr")) {
		unsigned long reg;
		unsigned long sv;
		if (ac < 4) {
			printf("mtgpr requires the GPR and a value\n");
			return 0;
		}
		if (av[idx+1][0] == 'r')
			reg = strtoul(&av[idx+1][1], 0, 0);
		else
			reg = strtoul(av[idx+1], 0, 0);
		idx++;
		sv = strtoul(av[idx+1], 0, 0);
		if (reg < 0 || reg > 31) {
			printf("mtgpr register must be valid (0 -> 31)\n");
			return 0;
		}
		context->gpr[reg] = sv;
		return 0;
	}
	if (!strcmp(av[idx], "mfgpr")) {
		unsigned long sv;
		int i;
		if (ac < 3) {
			printf("mfgpr requires the GPR(s)\n");
			return 0;
		}
		i = idx+1;
		while (i < ac) {
		    if (av[i][0] == 'r')
			    sv = strtoul(&av[i][1], 0, 0);
		    else
			    sv = strtoul(av[i], 0, 0);
		    if (sv < 0 || sv > 31) {
			    printf("mfgpr register must be valid (0 -> 31)\n");
			    return 0;
		    }
		    printf("GPR %d = 0x%x\n", sv, context->gpr[sv]);
		    i++;
		}
		return 0;
	}
	if (!strcmp(av[idx], "bt")) {
		backtrace(context->gpr[1]);
		return 0;
	}
	printf("unknown task command: %s\n", av[idx]);
	return 0;
}

int
xdis(int ac, char **av, struct pt_regs *regs) {
	static unsigned long adrs = 0x80000000;
	static unsigned long ni = 16;
	if (ac >= 2) {
		if (!strcmp(av[1], "pc"))
			adrs = regs->nip;
		else if (!strcmp(av[1], "lr"))
			adrs = regs->link;
		else
			adrs = strtoul(av[1], 0, 0);
	}
	if (ac >= 3) 
		ni = strtoul(av[2], 0, 0);
	adrs += ppc_inst_dump(adrs, ni);
	return 0;
}

int
xdump(int ac, char **av, struct pt_regs *regs) {
	static unsigned long adrs = 0x80000000;
	static unsigned long ni = 64;
	if (ac == 1) {
		prdump(adrs, ni);
		adrs += ni;
		return 0;
	}
	if (ac == 2) {
		adrs = strtoul(av[1], 0, 0);
		prdump(adrs, ni);
		adrs += ni;
		return 0;
	}
	adrs = strtoul(av[1], 0, 0);
	ni = strtoul(av[2], 0, 0);
	prdump(adrs, ni);
	adrs += ni;
	return 0;
}

int
xbt(int ac, char **av, struct pt_regs *regs) {
	if (ac == 1)
		backtrace(0);
	else {
		unsigned int sp = strtoul(av[1], 0, 0);
		backtrace(sp);
	}
	return 0;
}

int
xss(int ac, char **av, struct pt_regs *regs) {
	singlestep = 1;
	return 1;
}

static int
_xcondition_task_userbit(struct pt_regs *regs, struct bpt *bp) {
	if (!(regs->msr & MSR_PR))
		return 0;
	else {
		printf("breakpoint task_userbit at %p satisfied\n", bp->address);
		return 1;
	}
}

static int
_xcondition_nip(struct pt_regs *regs, struct bpt *bp) {
	if (bp->cond_data == regs->nip) {
		printf("breakpoint nip %p at %p satisfied\n", bp->cond_data, bp->address);
		return 1;
	}
	return 0;
}

// For kernel self-debugging use.
// size: 0=byte, 1=short, 2=int, 3=cache line
// rw: 0=read, 1=write, 2=read/write
unsigned int xSetDataBp(unsigned int addr, int size, int rw) 
{
    struct pt_regs regs;
    int which;

    // No slot for it, return illegal/unlikely address (failure).
    if (dabr[0].enabled && dabr[1].enabled) {
        return 1;
    }

    if (dabr[0].enabled) {
        which = 1;
    } else {
        which = 0;
    }
    dabr[which].enabled = 1;
    dabr[which].address = addr;
    dabr[which].rw = rw;
    dabr[which].size = size;

    // Insert w/ fake regs structure.
    regs.nip = 1;
    remove_bpts(&regs);	// not called through xmon, must keep assumptions
    insert_bpts(&regs);

    return addr;
}

// For kernel self-debugging use
void xRemoveDataBp(unsigned int addr)
{
    int which;
    struct pt_regs regs;

    if (dabr[0].enabled && dabr[0].address == addr) {
        which = 0;
    } else if (dabr[1].enabled && dabr[1].address == addr) {
        which = 1;
    } else {
        return;
    }

    dabr[which].enabled = 0;

    // Insert w/ fake regs structure.
    regs.nip = 1;
    remove_bpts(&regs);	// not called through xmon, must keep assumptions
    insert_bpts(&regs);
}

int
xbp(int ac, char **av, struct pt_regs *regs) {
	struct bpt *bp;
	int which;
	int i;
	unsigned int addr;
	if (ac == 1) {
		if (dabr[0].enabled)
			printf("0x%x: data comparator 1 for %s, size %s\n",
				dabr[0].address,
				(dabr[0].rw == 0 ? "R" : dabr[0].rw == 1 ? "W" : "RW"),
				(dabr[0].size == 0 ? "byte" : dabr[0].size == 1 ? "halfword" :
					dabr[0].size == 2 ? "word" : "quadword"));
		if (dabr[1].enabled)
			printf("0x%x: data comparator 2 for %s, size %s\n",
				dabr[1].address,
				(dabr[1].rw == 0 ? "R" : dabr[1].rw == 1 ? "W" : "RW"),
				(dabr[1].size == 0 ? "byte" : dabr[1].size == 1 ? "halfword" :
					dabr[1].size == 2 ? "word" : "quadword"));
		if (iabr[0].enabled)
			printf("0x%x: instruction comparator 1\n", iabr[0].address);
		if (iabr[1].enabled)
			printf("0x%x: instruction comparator 2\n", iabr[1].address);
		bp = bpts;
		for (i = 0; i < NBPTS; ++i, ++bp)
			if (bp->enabled)
				printf("0x%x: oinst(0x%x)\n", bp->address, bp->instr);
		return 0;
	}
	if (!strcmp(av[1], "set")) {
		int size = 0;
		int rw = 0;
		if (ac < 3) {
			printf("\"bp set\" requires an address or comparator\n");
			return 0;
		}
		if (!strcmp(av[2], "ihw")) {
			unsigned int addr;
			if (ac < 4) {
				printf("address required\n");
				return 0;
			}
			if (iabr[0].enabled && iabr[1].enabled) {
				printf("no instruction comparators available\n");
				return 0;
			}
			addr = strtoul(av[3], 0, 0);
			if (!iabr[0].enabled) {
				iabr[0].enabled = 1;
				iabr[0].address = addr;
				which = 1;
			} else {
				iabr[1].enabled = 1;
				iabr[1].address = addr;
				which = 2;
			}
			printf("breakpoint set in comparator %d\n", which);
			return 0;
		}
		if (!strcmp(av[2], "cond") ||
		    !strcmp(av[2], "next") ||
		    (av[2][0] >= '0' && av[2][0] <= '9')) {
			int dad = 0;
			int (*cond)(struct pt_regs *, struct bpt *) = 0;
			unsigned cond_data = 0;
			int idx = 2;
			if (!strcmp(av[idx], "next")) {
				if (ac < (idx + 2)) {
					printf("insufficient arguments\n");
					return 0;
				}
				addr = strtoul(av[idx + 1], 0, 0);
				dad = 1;
			}
			else if (!strcmp(av[idx], "cond")) {
				if (ac < idx + 3) {
					printf("insufficient arguments\n");
					return 0;
				}
				if (!strcmp(av[idx + 1], "task_userbit"))
					cond = _xcondition_task_userbit;
				if (!strcmp(av[idx + 1], "nip")) {
					if (ac < idx + 4) {
						printf("insufficient arguments\n");
						return 0;
					}
					cond = _xcondition_nip;
					cond_data = strtoul(av[idx+2], 0, 0);
					idx++;
				}
				else {
					printf("unknown condition\n");
					return 0;
				}
				addr = strtoul(av[idx + 2], 0, 0);
			}
			else
				addr = strtoul(av[idx], 0, 0);
			bp = bpts;
			for (i = 0; i < NBPTS; ++i, ++bp)
				if (!bp->enabled) {
					bp->enabled = 1;
					bp->address = addr;
					bp->count = 0;
					bp->sstep = 0;
					bp->condition = cond;
					bp->cond_data = cond_data;
					if (dad)
						bp->autodel = 1;
					break;
				}
			if (i == NBPTS) 
				printf("no free breakpoints available\n");
			return 0;
		}
		if (ac < 4) {
			printf("insufficient arguments to command\n");
			return 0;
		}
		if (dabr[0].enabled && dabr[1].enabled) {
			printf("no data comparators available\n");
			return 0;
		}
		if (!strcmp(av[2], "dacr")) {
			addr = strtoul(av[3], 0, 0);
			rw = 0;
		}
		else if (!strcmp(av[2], "dacw")) {
			addr = strtoul(av[3], 0, 0);
			rw = 1;
		}
		else if (!strcmp(av[2], "dac")) {
			addr = strtoul(av[3], 0, 0);
			rw = 2;
		}
		else {
			printf("unrecognized comparitor type\n");
			return 0;
		}
		if (ac >= 5) {
			if (!(strcmp(av[4], "byte")))
				size = 0;
			else if (!(strcmp(av[4], "half")))
				size = 1;
			else if (!(strcmp(av[4], "word")))
				size = 2;
			else if (!(strcmp(av[4], "quad")))
				size = 3;
			else {
				printf("unrecognized size qualifier\n");
				return 0;
			}
		}
		if (!dabr[0].enabled) {
			dabr[0].enabled = 1;
			dabr[0].address = addr;
			dabr[0].rw = rw;
			dabr[0].size = size;
			which = 1;
		} else {
			dabr[1].enabled = 1;
			dabr[1].address = addr;
			dabr[1].rw = rw;
			dabr[1].size = size;
			which = 2;
		}
		printf("breakpoint set in comparator %d\n", which);
		return 0;
	}
	if (!strcmp(av[1], "clear")) {
		if (ac < 3) {
			printf("command needs a breakpoint to clear\n");
			return 0;
		}
		if (av[2][0] >= '0' && av[2][0] <= '9') {
			addr = strtoul(av[2], 0, 0);
			bp = bpts;
			for (i = 0; i < NBPTS; ++i, ++bp)
				if (bp->enabled && bp->address == addr) {
					bp->enabled = 0;
					break;
				}
			if (i == NBPTS) 
				printf("no breakpoint found to clear\n");
			return 0;
		}
		if (ac < 4) {
			printf("command requires a comparator number\n");
			return 0;
		}
		which = strtoul(av[3], 0, 0);
		if (which != 1 && which != 2) {
			printf("invalid comparator number\n");
			return 0;
		}
		which--;
		if (!strcmp(av[2], "ihw"))
			iabr[which].enabled = 0;
		else if (!strcmp(av[2], "dac"))
			dabr[which].enabled = 0;
		else {
			printf("unknown breakpoint type\n");
			return 0;
		}
	}
	return 0;

}

void
xtlbdump(int first, int last, unsigned long search) {
	unsigned int opid = mfspr(SPRN_PID);
	if (search)
		printf("searching for 0x%x in tlb ...\n", search);
	printf(" I  EPN      S V PID    RPN      X W ZS WIMG\n");
	for (; first <= last; first++) {
		unsigned long tlbhi;
		unsigned long tlblo;
		asm volatile("tlbre %0,%1,0" : "=r" (tlbhi) : "r" (first));
		asm volatile("tlbre %0,%1,1" : "=r" (tlblo) : "r" (first));
		if ((search && ((search & 0xfffff000) == (tlbhi & 0xfffff000))) ||
		    !search) {
		    printf("%2d: %8x %d %d %3d -> %8x %d %d %2d %x\n",
			    first, (tlbhi & 0xfffffc00), ((tlbhi & 0x380) >> 7),
			    ((tlbhi & 0x40) ? 1 : 0), mfspr(SPRN_PID),
			    (tlblo & 0xfffffc00), ((tlblo & 0x200) ? 1 : 0),
			    ((tlblo & 0x100) ? 1 : 0), ((tlblo & 0xf0) >> 4),
			    (tlblo & 0xf));
		}
	}
	mtspr(SPRN_PID, opid);
}

int
xtlb(int ac, char **av, struct pt_regs *regs) {
	static unsigned int first = 0;
	static unsigned int last = 63;
	unsigned long addr = 0;
	int aidx = 1;
	if (!strcmp(av[0], "tsearch")) {
		if (aidx == ac) {
			printf("insufficient arguments\n");
			return 0;
		}
		addr = strtoul(av[aidx], 0, 0);
		aidx++;
	}
	if (ac == aidx) {
		xtlbdump(first, last, addr);
		return 0;
	}
	first = strtoul(av[aidx], 0, 0);
	aidx++;
	if (ac == aidx) {
		xtlbdump(first, last, addr);
		return 0;
	}
	last = strtoul(av[aidx], 0, 0);
	xtlbdump(first, last, addr);
	return 0;
}

int
xreset(int ac, char **av, struct pt_regs *regs) {
	printf("Resetting the system; Goodbye!\n");
	udelay(100000);
	/* Do a reset (make sure 5505 are properly reset) */
	hard_reset_now(); /* call from arch/ppc/tcd/setup.c */
	return 0;
}

int
xstop(int ac, char **av, struct pt_regs *regs) {
	int i;
	unsigned int trap;
	if (ac == 1) {
		for (i = 0; _exc_table[i].trap != 0; i++) {
			if (_exc_table[i].stop) {
                            if (_exc_table[i].stop == -1)
				printf("** %s (all)\n", _exc_table[i].name);
                            else
				printf("** %s (%d)\n", _exc_table[i].name,
                                        _exc_table[i].stop);
                        }
                        else
				printf("-- %s\n", _exc_table[i].name);
                }
		return 0;
	}
	if (ac < 3) {
		printf("unsufficient arguments\n");
		return 0;
	}
        for (i = 0; _exc_table[i].trap != 0; i++) {
            if (!strcmp(_exc_table[i].name, av[2])) {
                break;
            }
        }
        if (!_exc_table[i].trap) {
            printf("unknown trap: %s\n", av[2]);
            return 0;
        }
        trap = i;
	if (!strcmp(av[1], "after")) {
		_xdebug_stopafter = trap;
		return 0;
	}
	if (strcmp(av[1], "set") && strcmp(av[1], "clear")) {
		printf("unknown stop command\n");
		return 0;
	}
        if (!strcmp(av[1], "set")) {
                if (ac == 4)
                        _exc_table[trap].stop = strtoul(av[3], 0, 0);
                else
                        _exc_table[trap].stop = -1;
        }
        else
                _exc_table[trap].stop = 0;
        return 0;
}

int
xnext(int ac, char **av, struct pt_regs *regs) {
	unsigned int instr;
	if (mread(regs->nip, &instr, 4) != 4) {
		printf("no valid instruction present\n");
		return 0;
	}
	if (ISSUBR(instr)) {
		char addr[16];
		char *av[4];
		sprintf(addr, "0x%lx", regs->nip + 4);
		av[0] = "bp";
		av[1] = "set";
		av[2] = "next";
		av[3] = addr;
		xbp(4, av, regs);
		return 1;
	}
	singlestep = 1;
	return 1;
}

int
xpatch(int ac, char **av, struct pt_regs *regs) {
	unsigned long addr;
	unsigned long value;
	int wlen = 4;
	if (ac < 3) {
		printf("insufficient arguments\n");
		return 0;
	}
	addr = strtoul(av[1], 0, 0);
	value = strtoul(av[2], 0, 0);
	if (ac > 3) {
		if (!strcmp(av[3], "word"))
			wlen = 4;
		else if (!strcmp(av[3], "short"))
		{
		    *(short*)&value = (short)value; // work swap
		    wlen = 2;
		}
		else if (!strcmp(av[3], "byte"))
		{
		    *(char*)&value = (char)value; // byte swap
		    wlen = 1;
		}
		else {
			printf("unknown patch size\n");
			return 0;
		}
	}
	if (mwrite(addr, (void *)&value, wlen) != wlen)
		printf("error writing 0x%lx to 0x%lx\n", value, addr);
	return 0;
}

int
xsl(int ac, char **av, struct pt_regs *regs) {
	unsigned long addr;
	unsigned long value;
	int wlen = 4;
	if (ac < 2) {
		printf("insufficient arguments\n");
		return 0;
	}
	addr = strtoul(av[1], 0, 0);
	value = strtoul(av[2], 0, 0);
#if 1
	*(long*)addr = value;
	store_inst( addr );
#else	
	if (mwrite(addr, (void *)&value, wlen) != wlen)
		printf("error writing 0x%lx to 0x%lx\n", value, addr);
#endif
	return 0;
}
int
xsw(int ac, char **av, struct pt_regs *regs) {
	unsigned long addr;
	unsigned short value;
	int wlen = 2;
	if (ac < 2) {
		printf("insufficient arguments\n");
		return 0;
	}
	addr = strtoul(av[1], 0, 0);
	value = strtoul(av[2], 0, 0);
#if 1
	*(short*)addr = value;
	store_inst( addr );
#else	
	if (mwrite(addr, (void *)&value, wlen) != wlen)
		printf("error writing 0x%lx to 0x%lx\n", value, addr);
#endif
	return 0;
}
int
xsb(int ac, char **av, struct pt_regs *regs) {
	unsigned long addr;
	unsigned char value;
	int wlen = 1;
	if (ac < 2) {
		printf("insufficient arguments\n");
		return 0;
	}
	addr = strtoul(av[1], 0, 0);
	value = strtoul(av[2], 0, 0);
	if (mwrite(addr, (void *)&value, wlen) != wlen)
		printf("error writing 0x%lx to 0x%lx\n", value, addr);
	return 0;
}

int
xshow(int ac, char **av, struct pt_regs *regs) {
	int i;
	int sc;
	if (ac < 2) {
            for (i = 0; _exc_table[i].trap != 0; i++) {
                if (_exc_table[i].report) {
                    if (_exc_table[i].report == -1)
                        printf("** %s\n", _exc_table[i].name);
                    else
                        printf("** %s (%d)\n", _exc_table[i].name, _exc_table[i].report);
                }
                else
                    printf("-- %s\n", _exc_table[i].name);
            }
            return 0;
	}
	if (strcmp(av[1], "set") && strcmp(av[1], "clear")) {
		printf("\"set\" or \"clear\" required\n");
		return 0;
	}
	if (ac < 3) {
		printf("insufficient arguments\n");
		return 0;
	}
	if (!strcmp(av[1], "set"))
		sc = -1;
	else
		sc = 0;
        for (i = 0; _exc_table[i].trap != 0; i++) {
            if (!strcmp(av[2], _exc_table[i].name)) {
                if (ac >= 4)
                    _exc_table[i].report = strtoul(av[3], 0, 0);
                else
                    _exc_table[i].report = sc;
                break;
            }
        }
        if (!_exc_table[i].trap)
            printf("unknown trap type: %s\n", av[2]);
	return 0;
}

int
xtrace(int ac, char **av, struct pt_regs *regs) {
	int nt;
	if (ac < 2) {
		printf("insufficient arguments\n");
		return 0;
	}
	nt = strtoul(av[1], 0, 0);
	if (nt > 0) {
		_xdebug_trace = nt;
		return 1;
	}
	printf("invalid trace count\n");
	return 0;
}

int
xvm(int ac, char **av, struct pt_regs *regs) {
	unsigned int msr;
	if (ac < 2) {
		printf("insufficient arguments\n");
		return 0;
	}
	if (!strcmp(av[1], "on")) {
		msr = mfmsr();
		msr |= MSR_IR|MSR_DR;
		mtmsr(msr);
	}
	else if (!strcmp(av[1], "off")) {
		msr = mfmsr();
		msr &= ~(MSR_IR|MSR_DR);
		mtmsr(msr);
	}
	else
		printf("invalid argument\n");
	return 0;
}

int
xcache(int ac, char **av, struct pt_regs *regs) {
	static int dc_start = 0;
	static int dc_end = 255;
	static int ic_start = 0;
	static int ic_end = 511;
	unsigned long p;
	unsigned long stop;
	unsigned long addr;
	int which;
	int index;
	int aidx;
	int search = 0;
	if (!strcmp(av[0], "icache"))
		which = 0;
	else if (!strcmp(av[0], "dcache"))
		which = 1;
	else if (!strcmp(av[0], "isearch")) {
		which = 0;
		search = 1;
	}
	else if (!strcmp(av[0], "dsearch")) {
		which = 1;
		search = 2;
	}
	else {
		printf("illegal argument\n");
		return 0;
	}
	aidx = 1;
        addr = 0;
	if (search) {
		if (ac < 2) {
			printf("insufficient arguments\n");
			return 0;
		}
		addr = strtoul(av[1], 0, 0);
		aidx++;
	}
	if (ac > aidx) {
		int sline = strtoul(av[aidx], 0, 0);
		if (sline < 0 || (which && sline >= 256) || (!which && sline >= 512)) {
			printf("invalid line number\n");
			return 0;
		}
		if (which)
			dc_start = dc_end = sline;
		else
			ic_start = ic_end = sline;
		aidx++;
	}
	if (ac > aidx) {
		int sline = strtoul(av[aidx], 0, 0);
		if (sline < 0 ||
		    (which && sline >= 256) ||
		    (!which && sline >= 512)) {
			printf("invalid line number\n");
			return 0;
		}
		if (which)
			dc_end = sline;
		else
			ic_end = sline;
		aidx++;
	}

	if (which) {
	    p = (KERNELBASE & PAGE_MASK) + (dc_start * 16);
	    stop = p + (dc_end * 16);
	    index = dc_start;
	    if (search)
		printf("data cache search for 0x%x from line %d to %d\n",
			addr, dc_start, dc_end);
	    else
		printf("data cache dump from line %d to %d\n", dc_start, dc_end);
    	}
	else {
	    p = (KERNELBASE & PAGE_MASK) + (ic_start * 16);
	    stop = p + (ic_end * 16);
	    index = ic_start;
	    if (search)
		printf("instruction cache search for 0x%x from line %d to %d\n",
			addr, ic_start, ic_end);
	    else
		printf("instruction cache dump from line %d to %d\n", ic_start, ic_end);
    	}
	/* for every line in the cache ... */
	for (; p < stop; p += 16, index++) {
		int i;
		unsigned int tag;
		/* for each way in the set ... */
		if (!search)
		    printf("%3d ", index);
		for (i = 0; i < 2; i++) {
			int pr = 0;
			mtspr(SPRN_CDBCR, (0x00000010 | i));
			if (which)
				asm("dcread %0,0,%1" : "=r" (tag) : "r" (p));
			else {
				asm("icread 0,%0;nop" : : "r" (p));
				tag = mfspr(SPRN_ICDBDR);
			}
			if ((search == 1 &&
			    (tag & 0xfffffc00) == (addr & 0xfffffc00)) ||
			    (search == 2 &&
			    (tag & 0xfffffc00) == (addr & 0xfffffc00))) {
			 	printf("Match at entry %d:\n", index);
			 	pr = 1;
			}
			else if (!search)
				pr = 1;
			if (pr) {
			    printf(" (%c) %08x %c %c %c  ",
			    (i ? 'B' : 'A'),
			    ((which ? (tag & 0xfffff000) : (tag & 0xfffffc00))) | (16 * index),
			    (tag & 0x00000010 ? 'V' : ' '),
			    (which ? (tag & 0x00000020 ? 'D' : ' ') : ' '),
			    (tag & 0x00000001 ? (i ? 'L' : ' ') : (i ? ' ' : 'L')));
			}
			if (search && pr)
				printf("\n");
		}
		if (!search)
		    printf("\n");
	}
	return 0;
}

#define MAX_ARGV	32
int xhelp(int, char **, struct pt_regs *);
static char *xargv[MAX_ARGV];
static int xargc = 0;
static struct {
	char	*cmd;
	int	(*proc)(int ac, char **av, struct pt_regs *);
	char	*help;
} cmdtable[] = {
{ "task",	xregister_context,   "task [bt|mfgpr GPR|mtgpr GPR VALUE ...]" },
{ "dis",	xdis, "dis [ADDRESS [COUNT]]" },
{ "dump",	xdump, "dump [ADDRESS [COUNT]]" },
{ "bt",		xbt, "bt" },
{ "s",		xss, "s" },
{ "n",		xnext,"n" },
{ "bp",		xbp, "bp [set [ADDRESS|ihw|dac[r|w] TYPE]|clear [ADDRESS|ihw WHICH|dac WHICH]]" },
{ "tlb",	xtlb, "tlb [first [last]]" },
{ "tsearch",	xtlb, "tsearch ADDRESS [start [end]]" },
{ "reset",	xreset, "reset" },
{ "stop",	xstop, "stop [set trapno|clear trapno]" },
{ "os",		xos, "os OSFUNC [args ...]" },
{ "vm",		xvm, "vm (on|off)" },
{ "icache",	xcache, "icache [start [end]]" },
{ "dcache",	xcache, "dcache [start [end]]" },
{ "isearch",	xcache, "isearch ADDRESS [start [end]]" },
{ "dsearch",	xcache, "dsearch ADDRESS [start [end]]" },
{ "show",	xshow, "show [(set|clear) [traps]+]" },
{ "patch",	xpatch, "patch ADDRESS VALUE [word|short|byte]" },
{ "sl",         xsl, "sl ADDRESS VALUE" },
{ "sw",         xsw, "sw ADDRESS VALUE" },
{ "sb",	        xsb, "sb ADDRESS VALUE" },
{ "help",	xhelp, "help" },
{ "mtspr",	xregister_immediate, "mfspr SPRN" },
{ "mfspr",	xregister_immediate, "mtspr SPRN VALUE" },
{ "mtdcr",	xregister_immediate, "mfdcr DCRN" },
{ "mfdcr",	xregister_immediate, "mtdcr DCRN VALUE" },
{ "mfmsr",	xregister_immediate, "mfmsr" },
{ "mtmsr",	xregister_immediate, "mtmsr VALUE" },
{ 0 }
};

int
xhelp(int ac, char **av, struct pt_regs *regs) {
	int i;
	for (i = 0; cmdtable[i].cmd; i++)
		printf("%s\n", cmdtable[i].help);
	printf("resume|c\n");
	return 0;
}

char xlastline[256];

void
xmon(struct pt_regs *excp) {
	_xdebug_inxmon++;
	printf("(interrupts %s,", (mfmsr() & MSR_EE ? "on" : "off"));
	printf(" VM %s)\n", (mfmsr() & (MSR_IR|MSR_DR) ? "on" : "off"));
	for (;;) {
		extern char *xmon_readcmd(void); char *line;
		char *s;
		int i;

		if (_xdebug_restore_mm)
		    	printf("[spid %d] ", current->mm->context);
		printf("tcdmon> ");
		line = xmon_readcmd();
		for (s = line; *s != '\0'; s++) {
			if (*s == '\r') {
				strcpy(s, s+1);
				continue;
			}
			if (*s != ' ' && *s != '\t' && *s != '\n')
				break;
		}
		if (*s != '\0') {
			strncpy(xlastline, line, 255);
			xlastline[255] = '\0';
			s = strtok(xlastline, " \t\n");
			xargc = 0;
			while (xargc < MAX_ARGV && s) {
				xargv[xargc++] = s;
				s = strtok(0, " \t\n");
			}
		}
	    	if (xargc == 0)
			continue;

		if (!strcmp(xargv[0], "resume") || !strcmp(xargv[0], "c")) {
			_xdebug_inxmon--;
			return;
		}
		for (i = 0; cmdtable[i].cmd != 0; i++) {
			if (!strcmp(cmdtable[i].cmd, xargv[0])) {
				if ((*cmdtable[i].proc)(xargc, xargv, excp)) {
					_xdebug_inxmon--;
					return;
				}
				break;
			}
		}
		if (!cmdtable[i].cmd) {
			printf("unknown command\n");
		}
	}
}

void
backtrace(unsigned int btsp) {
	unsigned sp;
	unsigned stack[2];
	struct pt_regs regs;
	extern int int_return, critical_return, syscall_return;
	extern int prohibit_return;

	if (btsp != 0)
		sp = btsp;
	else
		sp = getsp();
	for (; sp != 0; sp = stack[0]) {
		if (mread(sp, stack, sizeof(stack)) != sizeof(stack)) {
			printf("(fault at %p)", sp);
			break;
		}
		printf("%x ", stack[1]);
		if (stack[1] == (unsigned) &int_return ||
		    stack[1] == (unsigned) &critical_return ||
		    stack[1] == (unsigned) &syscall_return ||
		    stack[1] == (unsigned) &prohibit_return) {
			if (mread(sp+16, &regs, sizeof(regs)) != sizeof(regs))
				break;
			printf("\nexception:%x [%x lr=%x] %x ", regs.trap, 
			       sp+16, regs.link, regs.nip);
			sp = regs.gpr[1];
			if (mread(sp, stack, sizeof(stack)) != sizeof(stack)) {
				printf("(fault at %p)", sp);
				break;
			}
		}
		else if (stack[0] <= sp) {
			if (stack[0] != 0)
				printf("(stack corrupted, next sp was %p)", stack[0]);
			break;
		}
	}
	printf("\n");
}

int
getsp()
{
    int x;

    asm("mr %0,1" : "=r" (x) :);
    return x;
}

void
excprint(struct pt_regs *fp)
{
	printf("vector: 0x%x at pc = %p, msr = 0x%x, sp = 0x%x [0x%x]\n",
		   fp->trap, fp->nip, fp->msr, fp->gpr[1], fp);
	switch (fp->trap) {
	case 0x700:	/* program check */
		printf("\tESR 0x%x\n", mfspr(SPRN_ESR));
		break;
	case 0x1100:	/* data TLB miss */
		printf("\tESR 0x%x DEAR 0x%x\n", mfspr(SPRN_ESR), mfspr(SPRN_DEAR));
		break;
	default:
	}
	if (current)
		printf("current = %p, pid = %d, comm = %s, regs = %p\n",
		       current, current->pid, current->comm, current->tss.regs);
}

void
prregs(struct pt_regs *fp)
{
	int n;

	for (n = 0; n < 32; ++n)
		printf("R%.2d = 0x%.8x%s", n, fp->gpr[n],
		       (n & 3) == 3? "\n": "   ");
	printf("pc  = 0x%.8x   msr = 0x%.8x   lr  = 0x%.8x   cr  = 0x%.8x\n",
	       fp->nip, fp->msr, fp->link, fp->ccr);
	printf("ctr = 0x%.8x   xer = 0x%.8x   trap = 0x%4x\n",
	       fp->ctr, fp->xer, fp->trap);
}

unsigned int
read_spr(int n)
{
    unsigned int instrs[2];
    int (*code)(void);

    instrs[0] = 0x7c6002a6 + ((n & 0x1F) << 16) + ((n & 0x3e0) << 6);
    instrs[1] = 0x4e800020;
    store_inst(instrs);
    store_inst(instrs+1);
    code = (int (*)(void)) instrs;
    return code();
}

void
write_spr(int n, unsigned int val)
{
    unsigned int instrs[2];
    int (*code)(unsigned int);

    instrs[0] = 0x7c6003a6 + ((n & 0x1F) << 16) + ((n & 0x3e0) << 6);
    instrs[1] = 0x4e800020;
    store_inst(instrs);
    store_inst(instrs+1);
    code = (int (*)(unsigned int)) instrs;
    code(val);
}

unsigned int
read_dcr(int n)
{
    unsigned int instrs[2];
    int (*code)(void);

    instrs[0] = 0x7c600286 + ((n & 0x1F) << 16) + ((n & 0x3e0) << 6);
    instrs[1] = 0x4e800020;
    store_inst(instrs);
    store_inst(instrs+1);
    code = (int (*)(void)) instrs;
    return code();
}

void
write_dcr(int n, unsigned int val)
{
    unsigned int instrs[2];
    int (*code)(unsigned int);

    instrs[0] = 0x7c600386 + ((n & 0x1F) << 16) + ((n & 0x3e0) << 6);
    instrs[1] = 0x4e800020;
    store_inst(instrs);
    store_inst(instrs+1);
    code = (int (*)(unsigned int)) instrs;
    code(val);
}

/*
 * Stuff for reading and writing memory safely
 */
int
mread(unsigned adrs, void *buf, int size)
{
	volatile int n;
	char *p, *q;

	n = 0;
	if( setjmp(bus_error_jmp) == 0 ){
		xmon_fault_handler = handle_fault;
		p = (char *) adrs;
		q = (char *) buf;
		switch (size) {
		case 2: *(short *)q = *(short *)p;	break;
		case 4: *(int *)q = *(int *)p;		break;
		default:
			for( ; n < size; ++n )
				*q++ = *p++;
		}
		n = size;
	}
	xmon_fault_handler = 0;
	return n;
}

int
mwrite(unsigned adrs, void *buf, int size)
{
	volatile int n;
	char *p, *q;

	n = 0;
	if( setjmp(bus_error_jmp) == 0 ){
		xmon_fault_handler = handle_fault;
		p = (char *) adrs;
		q = (char *) buf;
		for (; n < size; ++n) {
			*p++ = *q++;
			store_inst(p - 1);
		}
		n = size;
	} else {
		printf("*** Error writing address %p\n", adrs + n);
	}
	xmon_fault_handler = 0;
	return n;
}

void
handle_fault(struct pt_regs *regs)
{
        memcpy(&_xdebug_faultregs, regs, sizeof(*regs));
	longjmp(bus_error_jmp, 1);
}


void
prdump(unsigned adrs, int ndump)
{
	register int n, m, c, r, nr;
	unsigned char temp[16];

	for( n = ndump; n > 0; ){
		printf("%.8x", adrs);
		putchar(' ');
		r = n < 16? n: 16;
		nr = mread(adrs, temp, r);
		adrs += nr;
		for( m = 0; m < r; ++m ){
			putchar((m & 3) == 0 && m > 0? '.': ' ');
			if( m < nr )
				printf("%.2x", temp[m]);
			else
				printf("**");
		}
		for(; m < 16; ++m )
			printf("   ");
		printf("  |");
		for( m = 0; m < r; ++m ){
			if( m < nr ){
				c = temp[m];
				putchar(' ' <= c && c <= '~'? c: '.');
			} else
				putchar(' ');
		}
		n -= r;
		for(; m < 16; ++m )
			putchar(' ');
		printf("|\n");
		if( nr < r )
			break;
	}
}

#define GETWORD(v)	(((v)[0] << 24) + ((v)[1] << 16) + ((v)[2] << 8) + (v)[3])

int
ppc_inst_dump(unsigned adr, int count)
{
	int nr, dotted;
	unsigned first_adr;
	unsigned long inst, last_inst;
	unsigned char val[4];

	dotted = 0;
        last_inst = 0;
	for (first_adr = adr; count > 0; --count, adr += 4){
		nr = mread(adr, val, 4);
		if( nr == 0 ){
			printf("%.8x  ********\n", adr);
			break;
		}
		inst = GETWORD(val);
		if (adr > first_adr && inst == last_inst) {
			if (!dotted) {
				printf(" ...\n");
				dotted = 1;
			}
			continue;
		}
		dotted = 0;
		last_inst = inst;
		printf("%.8x  ", adr);
		printf("%.8x\t", inst);
		print_insn_big_powerpc(stdout, inst, adr);	/* always returns 4 */
		printf("\n");
	}
	return adr - first_adr;
}

void
print_address(addr)
unsigned addr;
{
	printf("%p", addr);
}

void
print_instr(unsigned int adr, unsigned int instr) {
	printf("%.8x\t", instr);
	print_insn_big_powerpc(stdout, instr, adr);	/* always returns 4 */
}
