/*
 *  linux/arch/ppc/kernel/process.h
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version
 *  2 of the License, or (at your option) any later version.
 *
 */

/*
 * This file contains information about how to build and interpret the process
 * log.
 */

#ifdef CONFIG_XMON
#define PLOG_CLKPERUS    (54000000 / 1000000)
#define PLOG_DIFF(cur, prev) \
        (((unsigned long) (cur) - (unsigned long) (prev)) / PLOG_CLKPERUS)

#define PLOG_TYPE_SCE	1
// "System call start"
struct _pl_sce {
    short pid;
    short scnum;
};

#define PLOG_TYPE_SCL	2
// "System call end"
struct _pl_scl {
    int pid;
    int rv;
};

#define PLOG_TYPE_RQ	3
// "Process added to run queue"
struct _pl_rq {
    short pid;
    short priority;
    short rt_priority;
    short policy;
    short pad; // it's happier if we're 4-byte aligned
    char fromintr;
    char preempt;
};

#define PLOG_TYPE_INTRE	4
// "Interrupt start"
#define PLOG_TYPE_INTRL	5
// "Interrupt end"

#define PLOG_TYPE_BHE	6
// "Bottom half start"
#define PLOG_TYPE_BHL	7
// "Bottom half end"

#define PLOG_TYPE_SWE	11
// "Process switch start"
struct _pl_swe {
    short frompid;
    short topid;
};

#define PLOG_TYPE_SWL 12
// "Process switch end"

#define PLOG_TYPE_SPS 13
// "Shared page at 0x%08x swapped out"

#define PLOG_TYPE_FMS 14
// "Data page ot 0x%08x (vaddr=0x%08x) swapped to backing file"

#define PLOG_TYPE_GPS 15
// "Data page at 0x%08x (vaddr=0x%08x) swapped out"

#define PLOG_TYPE_RESWAP 16
// "Data page at 0x%08x (vaddr=0x%08x) tossed out"

#define PLOG_TYPE_PAGEOUT 17
// "Memory at 0x%08x (vaddr=0x%08x) tossed out"

#define PLOG_TYPE_PAGEIN 18
// "Memory at vaddr 0x%08x paged in"

#define PLOG_TYPE_DEMAND_ZERO 19
// "Memory at vaddr 0x%08x demand zeroed"

#define PLOG_TYPE_PAGE_DIRTY 20
// "PTE 0x%08x being marked dirty"

#define PLOG_TYPE_NO_SWAP_PAGE 21
// "No swap page available to swap out pid %d, vaddr 0x%08x"

#define PLOG_TYPE_SHARED_PAGE_IGNORED 22
// "Shared page for pid %d, vaddr 0x%08x multiply referenced, not paged"

#define PLOG_TYPE_DIRTY_MULTIPLE_NO_SWAP 23
// "Page for pid %d, vaddr 0x%08x not swapped out -- it's being DMAed"


#define PLOG_TYPE_DRQ 25
// "marked unrunnable"
struct _pl_drq {
    short pid;
    short priority;
    short rt_priority;
    short policy;
};

extern void _pl_generic_event(int event, int somedata);
extern void _pl_generic_event2(int event, int data1, int data2);
extern void _pl_shared_page_swap(int virtPage);
extern void _pl_filemap_page_swap(int pid, int physPage, int virtPage);
extern void _pl_generic_page_swap(int pid, int physPage, int virtPage);
extern void _pl_reswap(int pid, int physPage, int virtPage);
extern void _pl_pageout(int pid, int physPage, int virtPage);
#endif // CONFIG_XMON
